;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Write a particular data set and its attributes.
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

function write_sds,sd_id,sds_name,sds_data,attr, $
                   start=start,count=count,compression=compression

;-------------------------------------------------------------------------------
; Write the sds corresponding to sds_name and all of its attributes.
;-------------------------------------------------------------------------------

result = size(sds_data)
ndims = result[0]
dims = result[1:ndims]
hdf_type = hdf_idl2hdftype(result[ndims+1])

natts = 0L
attr_names = ''
if (size(attr,/tname) eq 'STRUCT') then begin
  natts = n_tags
  attr_names = tag_names(attr)
endif

if (n_elements(start) eq ndims) then begin
  data_start = (start > 0) < (dims - 1)
endif else begin
  data_start = replicate(0L,ndims)
endelse

if (n_elements(count) eq ndims) then begin
  data_count = (count > 1) < (dims - start)
endif else begin
  data_count = dims
endelse

data_edge = (data_start + data_count) - 1L

CASE ndims OF
  1: sds_data_sub = sds_data[data_start[0]:data_edge[0]]
  2: sds_data_sub = sds_data[data_start[0]:data_edge[0],data_start[1]:data_edge[1]]
  3: sds_data_sub = sds_data[data_start[0]:data_edge[0],data_start[1]:data_edge[1],data_start[2]:data_edge[2]]
  4: sds_data_sub = sds_data[data_start[0]:data_edge[0],data_start[1]:data_edge[1],data_start[2]:data_edge[2],data_start[3]:data_edge[3]]
ELSE: $
  message,'Too many dimensions in SDS, 4 is the max allowed.'
ENDCASE

sds_id = hdf_sd_create(sd_id,sds_name,data_count,hdf_type=hdf_type)

;for n=0L, ndims-1L do begin
;  dimid  = hdf_sd_dimgetid(sds_id,n)
;  hdf_sd_dimset,dimid,name='nlon' + string(format='(I0)',n) 
;endfor

if (keyword_set(compression)) then hdf_sd_setcompress,sds_id,4,effort=5
hdf_sd_adddata,sds_id,sds_data_sub

if (natts gt 0) then begin
  for n=0L, natts-1L do begin
    hdf_sd_attrset,sds_id,attr_names[n],attr.(n)
  endfor
endif

hdf_sd_endaccess,sds_id

return,0
end

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

function hdf_sds_writer,     filename,                        $
                             data,                            $
                             path=path,                       $
			     start=start,                     $
			     count=count,                     $
			     hdf_struct=hdf_struct,           $
			     sds_index=sds_index,             $
			     attr=attr,                       $
			     scale_fct=scale_fct,             $
			     append_sds=append_sds,           $
			     compress=compress,               $
			     quiet=quiet,                     $
			     help=help
			     
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; NAME:
;       HDF_SDS_WRITER
;	
; PURPOSE:
;       This is a flexible routine for writing HDF SDS data sets and their
;       associated attributes (including global attributes)
;
; CATEGORY:
;        File writer
;
; CALLING SEQUENCE:
;        status = hdf_sds_writer(filename, $
;                                data, $
;                                path=path, $
;                                start=start, $
;                                count=count, $
;                                hdf_struct=hdf_struct, $
;                                sds_index=sds_index, $
;                                attr=attr, $
;                                scale_fct=scale_fct, $
;                                append_sds, $
;                                quiet=quiet, $
;                                help=help
;
; INPUTS:
;        FILENAME        The name of the HDF file to be created or appended on to
;
; OUTPUTS:
;        DATA            A structure variable containing all of the SDS's
;                        to be written.  The name of each structure member will
;                        be the same as in the hdf file.  For example: 
;                        The SDS "Radiance" is stored as data.radiance.
;                        Note that if hdf_struct is set, the SDS name tag
;                        is taken from the alias list, not the actual SDS
;                        name.  See keyword HDF_STRUCT for more details.
;
; RETURNS:
;        0: if successful
;       -1: if unsuccessful
;
; KEYWORD PARAMETERS:
;        PATH            Set this keyword to the name of the directory where 
;                        the hdf file is located.  The default is './'
;
;        START           Set this keyword to a vector containing the start 
;                        position in the array (s) to be written.  The default
;                        is [0,0,....0]
;
;        COUNT           Set this keyword to a vector containing the number
;                        of items in the array (s) to be written.  The default
;                        is to read all items.
;
;        HDF_STRUCT      Set this keyword to a 2 X N_sds string array so that
;                        hdf_struct = [[actual_sds_name1, alias_sds_name1], $
;                                      [actual_sds_name2, alias_sds_name2], ...]
;
;        SDS_INDEX       Set this keyword to a vector of SDS data indices.
;                        If hdf_struct is not set, then only SDS's with these
;                        index values in filename will be written.  If hdf_struct
;                        is set, only hdf_struct[*,sds_index] will be written.
;
;        ATTR            Set this keyword to a named structure variable in
;                        which all global attributes and attributes associated
;                        with each SDS are stored.
;                        The name of each structure member will be the same as
;                        in the hdf file.  For example: the global attribute
;                        "CREATED" is stored as attr.created.  The attrubute
;                        "units" associated with SDS name "Radiance" is
;                        stored as attr.radiance.units.  Note that if hdf_struct
;                        is set, the SDS name tag is taken from the alias list.
;
;        SCALE_FCT       Set this keyword to the name of a function (in string
;                        form) of type: result = scale_fct_name(scaled_data_array,
;                        attr) where scaled_data_array is an SDS read from the
;                        HDF file and attr is the same as ATTR.  This function is
;                        used to unscale scaled data in hdf_sds_reader prior to
;                        returning to the caller.
;
;        APPEND_SDS      Set this keyword to append onto an existing HDF file.
;
;        COMPRESS        Set this keyword if each SDS is to be compressed
;
;        QUIET           Set this keyword to avoid printing information about
;                        which SDS's are written.
;
;        HELP            Set this keyword to display USAGE information.
;
; MODIFICATION HISTORY
;        Written by:     Mike Pavolonis, NOAA/NESDIS, 19 June 2007
;
; EXAMPLES
;        
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

if ((n_params() ne 2) or $   
    (n_elements(filename) eq 0) or $
    (keyword_set(help))) then begin
  print,'USAGE:'
  print,'result = hdf_sds_writer(filename,'
  print,'                        data,'
  print,'                        path=path'
  print,'                        start=start'
  print,'                        count=count'
  print,'                        hdf_stuct=hdf_struct'
  print,'                        sds_index=sds_index'
  print,'                        attr=attr'
  print,'                        scale_fct=scale_fct'
  print,'                        append_sds=append_sds'
  print,'                        compress=compress'
  print,'                        quiet=quiet'
  print,'                        help=help'
  return,-1
endif

;-------------------------------------------------------------------------------
; Set the default file path.
;-------------------------------------------------------------------------------

if (keyword_set(path) eq 0) then path = './'

;-------------------------------------------------------------------------------
; Make sure the input file is HDF is the append_sds keyword is set.
;-------------------------------------------------------------------------------

if (keyword_set(append_sds)) then begin
  if not hdf_ishdf(path + filename) then message, 'Input parameter FILENAME is not HDF'
endif

;-------------------------------------------------------------------------------
; Open the HDF file in SDS write or read/write mode.
;-------------------------------------------------------------------------------

if (keyword_set(append_sds)) then begin
  sd_id = hdf_sd_start(path + filename,/rdwr)
endif else begin
  sd_id = hdf_sd_start(path + filename,/create)
endelse

;-------------------------------------------------------------------------------
; Determine how many data sets and global attributes are to be written, then
; write them.
;-------------------------------------------------------------------------------

nglobal_attr = 0L
nattr_total = 0L
if (n_elements(attr) gt 0) then begin
  if (size(attr,/tname) ne 'STRUCT') then begin
    message, 'Input keyword ATTR must be a structure.'
  endif
  nattr_total = n_tags(attr)
  attr_names = tag_names(attr)
  for n=0L, nattr_total-1L do begin
    if (size(attr.(n),/tname) ne 'STRUCT') then begin
      hdf_sd_attrset,sd_id,attr_names[n],attr.(n)
      nglobal_attr = nglobal_attr + 1L
    endif
  endfor
endif
nattr_sds = nattr_total - nglobal_attr

;-------------------------------------------------------------------------------
; If keyword hdf_struct is not set load it up with every sds name in the 
; data structure, otherwise, check to make sure that hdf_struct is of the
; expected format.
;-------------------------------------------------------------------------------

nsds_file = n_tags(data)
data_names = tag_names(data)

if (n_elements(hdf_struct) eq 0) then begin
  hdf_struct = strarr(2,nsds_file)
  nsds = nsds_file
  hdf_struct[0,*] = data_names
  hdf_struct[1,*] = data_names
endif else begin
  result = size(hdf_struct,/n_dimensions)
  if (result ne 2) then $
    message,'hdf_struct must be a 2 x N string array.'
  result = size(hdf_struct,/dimen)
  if (result[0] ne 2) then $
    message,'hdf_struct must be a 2 x N string array.'
  nsds = n_elements(hdf_struct[0,*])
endelse

;-------------------------------------------------------------------------------
; If a subset of the sds's listed in hdf_struct are to be written make sure
; keyword sds_index is set properly.
;-------------------------------------------------------------------------------

if (n_elements(sds_index) gt 0) then begin
  if (min(sds_index) lt 0 or max(sds_index) ge nsds) then begin
    print,format='("SDS index values must be 0 <= sds_index < ",I0)',nsds
    return,-1
  endif
  index = sds_index
endif else begin
  index = indgen(nsds)
endelse

;-------------------------------------------------------------------------------
; Determine how many sds's are to be read.
;-------------------------------------------------------------------------------

nsds_write = n_elements(index) 

;-------------------------------------------------------------------------------
; If keyword quiet is not set, print out a list of the sds's that will be written.
;-------------------------------------------------------------------------------

if (not keyword_set(quiet)) then begin
  print,''
  print,'Writing the following SDS(s): to ' + filename
  for n=0L, nsds_write-1L do $
      print,format='("SDS index = ",I3,2X,A0)',index[n],hdf_struct[0,index[n]]
  print,''
endif

;-------------------------------------------------------------------------------
; Loop over each sds that is to be written, write the data and attributes,
; apply the scale function if keyword scale_fct is set.
;-------------------------------------------------------------------------------

for n=0L, nsds_write-1L do begin
  
  ;-------------------------------------------------------------------------------
  ; Find the first SDS to be written.
  ;-------------------------------------------------------------------------------
  
  isds = where(data_names eq hdf_struct[0,index[n]],isavailable)
  if (isavailable eq 0) then begin
    print,'Cannot find ' + hdf_struct[0,index[n]] + ' in data structure.'
    continue
  endif
  sds_data = data.(isds)
  if (size(sds_data,/tname) eq 'STRUCT') then begin
    print,'SDS ' + hdf_struct[0,index[n]] + ' is not a valid data type.'
    continue
  endif
  
  ;-------------------------------------------------------------------------------
  ; If attributes exist for the current SDS, grab them.
  ;-------------------------------------------------------------------------------
  
  sds_attr = ''
  if (nattr_sds gt 0) then begin  
    
    isds = where(attr_names eq hdf_struct[0,index[n]],isavailable)
    if (isavailable eq 0) then sds_attr = attr.(isds)
    if (size(sds_attr,/tname) ne 'STRUCT') then begin
      print,'Attributes for a given SDS must be stored in a structure.'
      sds_attr = ''
      isavailable = 0L
    endif
    
    ;-------------------------------------------------------------------------------
    ; Apply the scale function if 'scale_fct' is set and the attributes are
    ; available.
    ;-------------------------------------------------------------------------------
      
    if (keyword_set(scale_fct) ne 0) then begin
      if (isavailable eq 0) then begin
        print,'Cannot find ' + hdf_struct[0,index[n]] + 'in attribute structure.'
      endif else begin
        sds_data = call_function(scale_fct,sds_data,sds_attr)
      endelse
    endif
    
  endif
  
  ;-------------------------------------------------------------------------------
  ; Write the current SDS and its attributes.
  ;-------------------------------------------------------------------------------
  
  status = write_sds(sd_id,hdf_struct[1,index[n]],sds_data, $
                     sds_attr,start=start, $
		     count=count,compression=keyword_set(compress))
endfor

;-------------------------------------------------------------------------------
; Close the hdf file.
;-------------------------------------------------------------------------------

hdf_sd_end,sd_id

;-------------------------------------------------------------------------------
; Return success.
;-------------------------------------------------------------------------------

return,0 
end
