function read_clavrx_hdf,filename,sds_name,$
                          area=area,noglobals=noglobals,$
                          gridcell_resolution=gridcell_resolution,$
gridcell_format=gridcell_format,$
units=units,hdf_filename=hdf_filename,$
start_year=start_year,end_year=end_year,$
start_day=start_day,end_day=end_day,$
start_time=start_time,end_time=end_time,$
c1=c1,c2=c2,solar_3b=solar_3b,a_3b=a_3b,b_3b=b_3b,nu_3b=nu_3b,$
a_4=a_4,b_4=b_4,nu_4=nu_4,a_5=a_5,b_5=b_5,nu_5=nu_5,$
sed=sed,ch1_gain_low=ch1_gain_low,ch1_gain_high=ch1_gain_high, $
ch1_switch_count=ch1_switch_count,ch1_dark_count=ch1_dark_count, $
quiet=quiet,temperature=temperature,actual_missing=actual_missing, $
time_error=time_error, $
error_msg = error_msg,stride= stride,_extra=_extra
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; This routine reads in CLAVR-x data (assuming the file is HDF).
;
; INPUT:
; filename: the full HDF filename [string]
; sds_name: the name of the HDF SDS [string]
;
; KEYWORDS:
; area: a 2 or 4-element vector indicating the area to be read
; [xo,nx] or [xo,yo,nx,ny]
; gridcell_resolution: returns the spatial resolution of
; the grid (-1 if pixel level file) [float]
; gridcell_format: returns the name of the grid
; projection ('' if pixel level file) [string]
; units: returns the physical units of the data [string]
; hdf_filename: returns the name of the hdf file read from
; (without the directory appended) [string]
; quiet: set to prevent printing to the screen
; temperature: set to convert radiances to brightness temps
; unscaled missing: return the unscaled missing value
;
; OUTPUT:
; unscaled: the unscaled data [intarr or fltarr]
;
; EXAMPLE:
; result = read_clavrx_hdf('NSS.GHRR.NJ.D98181.S2216.E2359.B1803839.GC.cell.hdf',$
; 'total_cloud_fraction')
;
; REQUIRED ROUTINES:
; NONE
;
; AUTHOR:
; Michael J. Pavolonis, CIMSS/UW-Madison
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ON_ERROR,2

MISSING = -999.0

;-----------------------------------------------------------------------------
; Check input.
;-----------------------------------------------------------------------------

if (n_params() ne 2) then $
  message,'USAGE: read_clavrx_hdf filename sds_name'
  
;-----------------------------------------------------------------------------
; Uncompress file if compressed with gzip.
;-----------------------------------------------------------------------------
filename_temp = filename
gzip_flg = 0
result = strmid(filename,strlen(filename)-3L,3)
if (result eq '.gz') then begin
  gzip_flg = 1
  filename_temp = strmid(filename,0,strlen(filename)-3L)
  spawn,'gunzip -cf ' + filename + ' > ' + filename_temp   ;uncompress without removed compressed file
  print, 'temp file = ', filename_temp
endif

;------------------------------------------------------------------------------
; Open HDF file for reading in SD mode.
;------------------------------------------------------------------------------

if (file_info(filename_temp)).size lt 3000 then begin
  message,'File size too small ' + filename_temp,/info
  if arg_present(error_msg) then error_msg = ' File size too small'
  
  return,-1
endif

error_read=''
catch,error_read
if error_read ne '' then begin
  message,'Error opening X ' + filename_temp,/info
  if arg_present(error_msg) then error_msg = ' error opening X'
 if n_elements(sd_id) gt 0 then HDF_SD_END,sd_id 
  return,-1

endif

sd_id = hdf_sd_start(filename_temp)
if (sd_id eq -1) then begin
  message,'Error opening ' + filename_temp,/info
  if arg_present(error_msg) then error_msg = ' error opening'
  if n_elements(sd_id) gt 0 then HDF_SD_END,sd_id 
  return,-1
endif
catch,/cancel
hdf_sd_fileinfo, sd_id, num_sds,num_attr
If (num_sds eq 0) then begin
  Message,'file empty ' + filename_temp,/info
  if arg_present(error_msg) then error_msg = ' file empty'
  if n_elements(sd_id) gt 0 then HDF_SD_END,sd_id 
  return,-1
Endif


;------------------------------------------------------------------------------
; Read in global attributes.
;------------------------------------------------------------------------------

attr_index = hdf_sd_attrfind(sd_id,'FILENAME')
if (attr_index eq -1) then message,'Error reading in FILENAME global attribute.'
hdf_sd_attrinfo,sd_id,attr_index,DATA=hdf_filename,count=count
hdf_filename = strmid(hdf_filename,0,strpos(hdf_filename,'.hdf')+4)

attr_index = hdf_sd_attrfind(sd_id,'START_YEAR')
if (attr_index eq -1) then message,'Error reading in START_YEAR global attribute.'
hdf_sd_attrinfo,sd_id,attr_index,DATA=start_year

attr_index = hdf_sd_attrfind(sd_id,'START_DAY')
if (attr_index eq -1) then message,'Error reading in START_DAY global attribute.'
hdf_sd_attrinfo,sd_id,attr_index,DATA=start_day

attr_index = hdf_sd_attrfind(sd_id,'START_TIME')
if (attr_index eq -1) then message,'Error reading in START_TIME global attribute.'
hdf_sd_attrinfo,sd_id,attr_index,DATA=start_time

attr_index = hdf_sd_attrfind(sd_id,'END_YEAR')
if (attr_index eq -1) then message,'Error reading in END_YEAR global attribute.'
hdf_sd_attrinfo,sd_id,attr_index,DATA=end_year

attr_index = hdf_sd_attrfind(sd_id,'END_DAY')
if (attr_index eq -1) then message,'Error reading in END_DAY global attribute.'
hdf_sd_attrinfo,sd_id,attr_index,DATA=end_day

attr_index = hdf_sd_attrfind(sd_id,'END_TIME')
if (attr_index eq -1) then message,'Error reading in END_TIME global attribute.'
hdf_sd_attrinfo,sd_id,attr_index,DATA=end_time

data_type = ''
attr_index = hdf_sd_attrfind(sd_id,'DATA_TYPE')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=data_type
endif

;--- read attributes for pixel level attributes
gridcell_resolution = -1
num_cells = -1
gridcell_resolution_unit = ''
if (data_type ne 'PIXEL') then begin

gridcell_format = ''
attr_index = hdf_sd_attrfind(sd_id,'GRIDCELL_FORMAT')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=gridcell_format
endif


attr_index = hdf_sd_attrfind(sd_id,'NUM_CELLS_WITH_DATA')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=num_cells
endif

attr_index = hdf_sd_attrfind(sd_id,'GRIDCELL_RESOLUTION')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=gridcell_resolution_tmp
  gridcell_resolution = gridcell_resolution_tmp[0]
endif

attr_index = hdf_sd_attrfind(sd_id,'GRIDCELL_RESOLUTION_UNIT')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=gridcell_resolution_unit
endif

endif

;---- begin akh mod 12/04
c1 = -1
attr_index = hdf_sd_attrfind(sd_id,'C1')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=c1
endif

c2 = -1
attr_index = hdf_sd_attrfind(sd_id,'C2')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=c2
endif

solar_3b= -1
attr_index = hdf_sd_attrfind(sd_id,'SOLAR_3B_NU')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=solar_3b
endif

a_3b = -1
attr_index = hdf_sd_attrfind(sd_id,'A_3B')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=a_3b
endif

b_3b = -1
attr_index = hdf_sd_attrfind(sd_id,'B_3B')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=b_3b
endif

nu_3b = -1
attr_index = hdf_sd_attrfind(sd_id,'NU_3B')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=nu_3b
endif

a_4 = -1
attr_index = hdf_sd_attrfind(sd_id,'A_4')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=a_4
endif

b_4 = -1
attr_index = hdf_sd_attrfind(sd_id,'B_4')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=b_4
endif

nu_4 = -1
attr_index = hdf_sd_attrfind(sd_id,'NU_4')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=nu_4 
endif

a_5 = -1
attr_index = hdf_sd_attrfind(sd_id,'A_5')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=a_5
endif

b_5 = -1
attr_index = hdf_sd_attrfind(sd_id,'B_5')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=b_5
endif

nu_5 = -1
attr_index = hdf_sd_attrfind(sd_id,'NU_5')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=nu_5 
endif

sed = -1
attr_index = hdf_sd_attrfind(sd_id,'SUN_EARTH_DISTANCE')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=sed
endif

sed = -1
attr_index = hdf_sd_attrfind(sd_id,'SUN_EARTH_DISTANCE')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=sed
endif

;time_error = -1
;attr_index = hdf_sd_attrfind(sd_id,'TIME_ERROR_SECONDS')
;if (attr_index ne -1) then begin
;  hdf_sd_attrinfo,sd_id,attr_index,DATA=time_error
;endif

ch1_gain_high = -1
attr_index = hdf_sd_attrfind(sd_id,'CH1_GAIN_HIGH')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=ch1_gain_high
endif

ch1_gain_low = -1
attr_index = hdf_sd_attrfind(sd_id,'CH1_GAIN_LOW')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=ch1_gain_low
endif

ch1_switch_count = -1
attr_index = hdf_sd_attrfind(sd_id,'CH1_SWITCH_COUNT')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=ch1_switch_count
endif

ch1_dark_count = -1
attr_index = hdf_sd_attrfind(sd_id,'CH1_DARK_COUNT')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,sd_id,attr_index,DATA=ch1_dark_count
endif


;---- end akh mods 12/04

if (keyword_set(quiet) eq 0) then begin
  print,''
  print,'Reading from ' + hdf_filename
  print,format='(A0,2X,2(I4,2X))','Starting/Ending Year:',start_year,end_year
  print,format='(A0,2X,2(I3,2X))','Starting/Ending Day:',start_day,end_day
  print,format='(A0,2X,2(F5.2,2X))','Starting/Ending Time:',start_time,end_time
  if (num_cells[0] ne -1 and data_type ne 'PIXEL') then begin
    print,format='(A0,2X,I0)','Total number of cells:',num_cells[0]
    print,'Gridcell resolution: ' + string(format='(F3.1)',gridcell_resolution[0]) + $
          ' ' + gridcell_resolution_unit
    print,format='(A0,2X,A0)','Gridcell format:',gridcell_format
  endif
endif

;-----------------------------------------------------------------------------
; Get the dimensions of the data to be read in.
;-----------------------------------------------------------------------------

var_index = hdf_sd_nametoindex(sd_id,sds_name)
if (var_index eq -1) then $
  message,'Selected SDS parameter is not available.'
varid = hdf_sd_select(sd_id,var_index)
   
hdf_sd_getinfo,varid,name=name,type=type,label=label,format=format,$
               hdf_type=scale_type,coordsys=coordsys,ndims=ndims,dims=dimen,$
natts=natts

;-----------------------------------------------------------------------------
; Determine the exact data range to be read in.
;-----------------------------------------------------------------------------

CASE ndims OF
  1: begin
    if (n_elements(area) eq 2) then begin
      if (keyword_set(quiet) eq 0) then print,'A subset of data is to be read.'
      start = lonarr(1)
      count = lonarr(1)
      start[0] = (long(area[0]) > 0L) < (dimen[0] - 1L)
      count[0] = (long(area[1]) > 1L) < (dimen[0] - start[0])
    endif else begin
;     if (keyword_set(quiet) eq 0) then print,'Full data array is to be read.'
      start = [0L]
      count = [dimen[0]]
    endelse
  end
  2: begin
    if (n_elements(area) eq 4) then begin
      if (keyword_set(quiet) eq 0) then print,'A subset of data is to be read.'
      start = lonarr(2)
      count = lonarr(2)
      start[0] = (long(area[0]) > 0L) < (dimen[0] - 1L)
      start[1] = (long(area[1]) > 0L) < (dimen[1] - 1L)
      count[0] = (long(area[2]) > 1L) < (dimen[0] - start[0])
      count[1] = (long(area[3]) > 1L) < (dimen[1] - start[1])
    endif else begin
;     print,'Full data array is to be read.'
      start = [0L,0L]
      count = [dimen[0],dimen[1]]
    endelse
  end
ELSE: message,sds_name + ' has too many dimensions.'
ENDCASE

;-----------------------------------------------------------------------------
; Read in selected data and unscale.
;-----------------------------------------------------------------------------

if (keyword_set(quiet) eq 0) then begin
  print,''
  print,'Reading in ' + sds_name + ' array...'
endif

hdf_sd_getdata,varid,data,start=start,count=count,stride=stride

attr_index = hdf_sd_attrfind(varid,'SCALED')
if (attr_index eq -1) then message,'Error reading in SCALED attribute.'
hdf_sd_attrinfo,varid,attr_index,DATA=scale_flg

;--- if this sds is scaled, read attributes required for unscaling
if (scale_flg[0] ge 1) then begin
  attr_index = hdf_sd_attrfind(varid,'RANGE_MIN')
  if (attr_index eq -1) then message,'Error reading in RANGE_MIN attribute.'
  hdf_sd_attrinfo,varid,attr_index,DATA=actual_min

  attr_index = hdf_sd_attrfind(varid,'RANGE_MAX')
  if (attr_index eq -1) then message,'Error reading in RANGE_MAX attribute.'
  hdf_sd_attrinfo,varid,attr_index,DATA=actual_max

  attr_index = hdf_sd_attrfind(varid,'RANGE_MISSING')
  if (attr_index eq -1) then message,'Error reading in RANGE_MISSING attribute.'
  hdf_sd_attrinfo,varid,attr_index,DATA=actual_missing

  attr_index = hdf_sd_attrfind(varid,'SCALED_MIN')
  if (attr_index eq -1) then message,'Error reading in SCALED_MIN attribute.'
  hdf_sd_attrinfo,varid,attr_index,DATA=scaled_min,hdf_type=s_type
 
  attr_index = hdf_sd_attrfind(varid,'SCALED_MAX')
  if (attr_index eq -1) then message,'Error reading in SCALED_MAX attribute.'
  hdf_sd_attrinfo,varid,attr_index,DATA=scaled_max,hdf_type=s_type
 
  if (s_type eq 'DFNT_INT8') then $
    message,'Scaled min/max is an 8 bit int. This is not acceptable.'

  attr_index = hdf_sd_attrfind(varid,'SCALED_MISSING')
  if (attr_index eq -1) then message,'Error reading in SCALED_MISSING attribute.'
  hdf_sd_attrinfo,varid,attr_index,DATA=scaled_missing

endif

attr_index = hdf_sd_attrfind(varid,'UNITS')
if (attr_index ne -1) then begin
  hdf_sd_attrinfo,varid,attr_index,DATA=units
endif else begin
  units = ''
endelse

if (units eq 'none') then begin
  units = '[unitless]'
endif else begin
  units = '[' + units + ']'
endelse

if (keyword_set(quiet) eq 0) then $
  print,'Converting scaled data to unscaled data in ' + units + '...'

;-----------------------------------------------------------------------------
; If scaled data is an 8 bit signed integer, IDL will wrap (-128) - (-1) to
; 128 - 255, this needs to be undone so that the scale factors and missing
; fill value and data array both range from -128 - 127.
;-----------------------------------------------------------------------------

if (scale_type[0] eq 'DFNT_INT8' and scale_flg[0] ge 1) then begin
  data = long(data)
  index = where(data ge 128L,nindex)
  if (nindex gt 0) then data[index] = data[index] - 256L
endif

if (scale_type[0] eq 'DFNT_INT8' and scale_flg[0] eq 0) then begin
  data = fix(data)
  index = where(data ge 128L,nindex)
  if (nindex gt 0) then data[index] = data[index] - 256S
endif

if (scale_flg[0] ge 1) then begin

  imiss = where(data eq scaled_missing[0],nmiss,complement=igood,ncomplement=ngood)
  
  unscaled = (float(data - scaled_min[0])/ $                        ; this varies from 0 to 1
              float(scaled_max[0] - scaled_min[0]))

  if (scale_flg[0] eq 1 or scale_flg[0] eq 2 ) then begin
    unscaled = actual_min[0] + (actual_max[0] - actual_min[0])*unscaled
  endif

  if (ngood gt 0 and scale_flg[0] eq 2) then unscaled[igood] = 10.0^unscaled[igood]

  if (ngood gt 0 and scale_flg[0] eq 3) then begin
    unscaled[igood] = actual_min[0] + (actual_max[0] - actual_min[0])*(unscaled[igood]^2)
  endif

  if (nmiss gt 0) then unscaled[imiss] = MISSING


; unscaled = (actual_min[0]) + (float(data - scaled_min[0])/ $
;                               float(scaled_max[0] - scaled_min[0]))* $
;                              (actual_max[0] - actual_min[0])
; imiss = where(data eq scaled_missing[0],nmiss,complement=igood,ncomplement=ngood)
; if (nmiss gt 0) then unscaled[imiss] = MISSING
; if (ngood gt 0 and scale_flg[0] eq 2) then unscaled[igood] = 10.0^unscaled[igood]
  
endif else begin

  unscaled = data
; imiss = where(data eq scaled_missing[0],nmiss,complement=igood,ncomplement=ngood)
  imiss = where(data eq -128,nmiss,complement=igood,ncomplement=ngood)
  if (nmiss gt 0) then unscaled[imiss] = MISSING
  
endelse

;-----------------------------------------------------------------------------
; Close the hdf file.
;-----------------------------------------------------------------------------

hdf_sd_endaccess,varid
hdf_sd_end,sd_id

if (keyword_set(quiet) eq 0) then $
  print,'SDS ' + sds_name + ' was successfully read in.'
  
;-----------------------------------------------------------------------------
; Gzip HDF file, if it was originally compressed.
;-----------------------------------------------------------------------------
;if (gzip_flg) then spawn,'gzip -f ' + filename_temp
 if (gzip_flg) then spawn,'rm ' + filename_temp


;---------------------------------------------------------------------------
; if temperature keyword set and sds_name is appropriate to a radiance then
; convert radiance to brightness temeperature
;---------------------------------------------------------------------------
if (keyword_set(temperature) ne 0) then  begin

 irad = 0
 if (strcmp(sds_name,'ch3b',4) eq 1) then begin
     irad = 1
     a = a_3b[0]
     b = b_3b[0]
     nu = nu_3b[0]
 endif
 if (strcmp(sds_name,'ch4',3) eq 1) then begin
     irad = 1
     a = a_4[0]
     b = b_4[0]
     nu = nu_4[0]
 endif
 if (strcmp(sds_name,'ch5',3) eq 1) then begin
     irad = 1
     a = a_5[0]
     b = b_5[0]
     nu = nu_5[0]
 endif
 
 if (irad eq 1) then begin
   index = where(unscaled ne MISSING,count)
   if (count gt 0) then begin
     unscaled[index] = a + b * ((c2[0]*nu) / alog( 1.0 + (c1[0]*(nu^3))/unscaled[index]))
   endif
  units = '[K]'
 endif

endif
  
return,unscaled
end

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
